﻿Imports MapInfo.Data
Imports System.IO
Imports MapInfo.Geometry
Imports System.Collections.Generic
Imports System.Data.Common

Namespace MapXtreme

    Public Class ShapefileFactory

        Private m_dataBaseName As String = Nothing
        Private m_userName As String = Nothing
        Private m_passWord As String = Nothing
        Private m_defaultExportFileName As String = "expshp"

        Public Sub New(ByVal dataBaseName As String, ByVal userName As String, ByVal passWord As String)
            m_dataBaseName = dataBaseName
            m_userName = userName
            m_passWord = passWord
        End Sub

        Public ReadOnly Property DefaultExportShapefileFileName() As String
            Get
                Return m_defaultExportFileName
            End Get
        End Property

        Public Function Export(ByVal oracleSqlToGet As String, ByVal columnsExportNames As List(Of String), ByVal typesExport As List(Of String), toUpperCase As Boolean, ByVal geometricalColumnName As String, ByVal pathToSavedFiles As String, Optional ByVal exportFileName As String = Nothing) As List(Of Feature)
            Dim myConnection As MIConnection = Nothing
            Dim ti As TableInfoServer = Nothing
            Dim myLocalTable As Table = Nothing
            Dim messageToThrow As String = Nothing
            Dim command As MICommand = Nothing
            Dim reader As MIDataReader = Nothing
            Dim creatorOfMainFile As MainAndIndexFileJuggler = Nothing
            Dim creatorOfDBFFile As DBFJuggler = Nothing
            Dim localFeature As Feature = Nothing
            Dim colsOfLocalFeature, colsFromReader As Columns
            Dim valuesOfAlphanumericData As List(Of Object)
            Dim stringValuesOfAlphanumericData As String
            Dim i, n As Integer
            Dim toReturn As List(Of Feature)
            Dim t As Table

            Try
                ' connexion au serveur Oracle
                myConnection = New MIConnection()

                ' si le nom du layer à un point, cela indique un schéma
                ' les points sont interdits dans les noms de layer en mémoire
                ti = New TableInfoServer("local_alias_" & Me.DefaultExportShapefileFileName)
                ti.ConnectString = "SRVR=" & m_dataBaseName & ";UID=" & m_userName & ";PWD=" & m_passWord & ";"
                ti.Query = oracleSqlToGet
                ti.Toolkit = ServerToolkit.Oci
                ti.CacheSettings.CacheType = CacheOption.Off

                myLocalTable = MapInfo.Engine.Session.Current.Catalog.OpenTable(ti)
                myConnection.Open()

                command = myConnection.CreateCommand()
                command.CommandText = "SELECT * FROM local_alias_" & Me.DefaultExportShapefileFileName
                reader = command.ExecuteReader()

                ' création fichier principal ESRI Shape
                creatorOfMainFile = New MainAndIndexFileJuggler(Me.DefaultExportShapefileFileName, pathToSavedFiles)

                ' création fichier DBF
                creatorOfDBFFile = New DBFJuggler(pathToSavedFiles)
                creatorOfDBFFile.CreateDBF(Me.DefaultExportShapefileFileName, columnsExportNames, typesExport)

                colsOfLocalFeature = New Columns()
                colsFromReader = reader.Columns

                ' Composer un objet "Columns", utilisé par la suite dans le construction des features
                n = reader.FieldCount - 3
                For i = 0 To colsFromReader.Count - 3 ' -3 pour eliminer les colonnes GEOM, ROWID et MY_STYLE
                    colsOfLocalFeature.Add(colsFromReader(i))
                    Console.WriteLine("colsFromReader({0}) = {1}", i, colsFromReader(i))
                Next

                valuesOfAlphanumericData = New List(Of Object)
                toReturn = New List(Of Feature)

                While reader.Read()
                    ' ----------------------------------------------------------------------------
                    ' données spatiales
                    ' ----------------------------------------------------------------------------
                    ' ajouter dans temporary_main
                    Dim featureGeometry = reader.GetFeatureGeometry(geometricalColumnName)
                    If Not creatorOfMainFile.WriteGeometryToFile(featureGeometry) Then
                        Continue While ' si l'insertion de géométrie n'a pas abouti, on passe directement au suivant, inutile de traiter les données alphanumériques
                    End If

                    ' ----------------------------------------------------------------------------
                    ' données alphanumériques
                    ' ----------------------------------------------------------------------------
                    valuesOfAlphanumericData.Clear()
                    stringValuesOfAlphanumericData = ""
                    For i = 0 To n - 1 ' n-1 parce que l'avant dernière colonne contient les Geometries et la derniere les ROWID
                        Dim value As Object = reader.GetValue(i)
                        If TypeOf value Is String AndAlso CStr(value).Length > 254 Then
                            value = CStr(value).Substring(0, 254)
                        End If
                        valuesOfAlphanumericData.Add(value)
                        If stringValuesOfAlphanumericData = "" Then
                            stringValuesOfAlphanumericData += "'"
                        Else
                            stringValuesOfAlphanumericData += "', '"
                        End If
                        stringValuesOfAlphanumericData += valuesOfAlphanumericData(i).ToString().Replace("'", " ")
                    Next
                    stringValuesOfAlphanumericData += "'"
                    creatorOfDBFFile.WriteToDbf(stringValuesOfAlphanumericData, toUpperCase)

                    ' ----------------------------------------------------------------------------
                    ' feature créé
                    ' ----------------------------------------------------------------------------
                    localFeature = New Feature(reader.Columns, valuesOfAlphanumericData.ToArray)
                    localFeature.Geometry = featureGeometry
                    toReturn.Add(localFeature)
                End While

                'former main à partir de temporary_main
                'former index à partir de temporary_index
                creatorOfDBFFile.CloseInternalConnection()
                creatorOfMainFile.FinaliseAndCloseFiles()
                reader.Close()
                command.Dispose()
                myConnection.Close()

                ' renommer les fichiers exportés dans le nom de fichier choisi si besoin
                If exportFileName IsNot Nothing Then
                    If File.Exists(pathToSavedFiles & "\" & exportFileName & ".shp") Then
                        File.Delete(pathToSavedFiles & "\" & exportFileName & ".shp")
                    End If
                    File.Move(pathToSavedFiles & "\" & Me.DefaultExportShapefileFileName & ".shp", pathToSavedFiles & "\" & exportFileName & ".shp")

                    If File.Exists(pathToSavedFiles & "\" & exportFileName & ".shx") Then
                        File.Delete(pathToSavedFiles & "\" & exportFileName & ".shx")
                    End If
                    File.Move(pathToSavedFiles & "\" & Me.DefaultExportShapefileFileName & ".shx", pathToSavedFiles & "\" & exportFileName & ".shx")

                    If File.Exists(pathToSavedFiles & "\" & exportFileName & ".dbf") Then
                        File.Delete(pathToSavedFiles & "\" & exportFileName & ".dbf")
                    End If
                    File.Move(pathToSavedFiles & "\" & Me.DefaultExportShapefileFileName & ".dbf", pathToSavedFiles & "\" & exportFileName & ".dbf")
                End If
            Catch ex As Exception
                messageToThrow = ex.Message & vbCrLf & "Query : " & ti.Query
                Throw New Exception(messageToThrow, ex)
            Finally
                t = MapInfo.Engine.Session.Current.Catalog.GetTable(ti.Alias)
                If Not t Is Nothing Then
                    MapInfo.Engine.Session.Current.Catalog.CloseTable(ti.Alias)
                End If
            End Try

            Return toReturn
        End Function

        Private Class DBFJuggler
            Private layerName As String = Nothing
            Private pathToDbfFile As String = Nothing
            Private dbConnection As OleDb.OleDbConnection = Nothing

            Public Sub New(ByVal pathToDbfFile As String)
                If pathToDbfFile Is Nothing Then
                    Me.pathToDbfFile = System.Environment.CurrentDirectory
                Else
                    Me.pathToDbfFile = pathToDbfFile
                End If

                ' Locale Identifier=1036; (french) ne change rien... : http://msdn.microsoft.com/en-us/library/office/aa140022(v=office.10).aspx
                Me.dbConnection = New OleDb.OleDbConnection("Provider=Microsoft.Jet.OLEDB.4.0;" & _
                    "Extended Properties=dBase IV;" & _
                    "Data Source=" & Me.pathToDbfFile & ";")
                Me.dbConnection.Open()
            End Sub

            Public Sub CreateDBF(ByVal tableName As String, ByVal alphaNumericalColumns As String)
                Dim localAlphaNumericalColumns As String = Nothing
                Dim stmt As DbCommand = Nothing
                Dim rs As DbDataReader = Nothing

                If tableName.Length > 8 Then
                    Me.layerName = tableName.Substring(0, 8).ToUpper()
                Else
                    Me.layerName = tableName
                End If
                localAlphaNumericalColumns = alphaNumericalColumns.Replace(",", " varchar,") & " varchar"
                stmt = dbConnection.CreateCommand()

                Try
                    stmt.CommandText = "DROP TABLE " & layerName.Replace(".", "_")
                    rs = stmt.ExecuteReader()
                Catch ex As Exception
                Finally
                    If Not rs Is Nothing Then
                        rs.Close()
                    End If
                End Try

                stmt.CommandText = "CREATE TABLE " & layerName.Replace(".", "_") & " (" + localAlphaNumericalColumns + ")"
                rs = stmt.ExecuteReader() 'si la table est read only, une excetion sera levée
                rs.Close()
            End Sub

            Public Sub CreateDBF(ByVal tableName As String, ByVal columnNames As List(Of String), ByVal columnTypes As List(Of String))
                Dim columns As String = Nothing
                Dim stmt As DbCommand = Nothing
                Dim rs As DbDataReader = Nothing
                Dim i As Integer = 0
                Dim column As String = Nothing

                If tableName.Length > 8 Then
                    Me.layerName = tableName.Substring(0, 8).ToUpper()
                Else
                    Me.layerName = tableName
                End If
                For i = 0 To columnNames.Count - 1
                    If columns Is Nothing Then
                        columns = columnNames(i) & " " & columnTypes(i)
                    Else
                        columns += ", " & columnNames(i) & " " & columnTypes(i)
                    End If
                Next
                stmt = dbConnection.CreateCommand()

                Try
                    stmt.CommandText = "DROP TABLE " & layerName.Replace(".", "_")
                    rs = stmt.ExecuteReader()
                Catch ex As Exception
                Finally
                    If Not rs Is Nothing Then
                        rs.Close()
                    End If
                End Try

                stmt.CommandText = "CREATE TABLE " & layerName.Replace(".", "_") & " (" + columns + ")"
                Console.WriteLine("CreateDBF = {0}", stmt.CommandText)
                rs = stmt.ExecuteReader() 'si la table est read only, une excetion sera levée
                rs.Close()
            End Sub

            Public Sub WriteToDbf(ByVal alphaNumericalValues As String, toUpperCase As Boolean)
                Dim stmt As DbCommand = Nothing
                Dim rs As DbDataReader = Nothing

                stmt = Me.dbConnection.CreateCommand()
                stmt.CommandText = Me.ReplaceAccent("INSERT INTO " & layerName.Replace(".", "_") & " VALUES (" & alphaNumericalValues & ")")

                If toUpperCase Then
                    stmt.CommandText = stmt.CommandText.ToUpper()
                End If

                Try
                    rs = stmt.ExecuteReader()
                Catch ex As Exception
                    Throw ex
                Finally
                    If Not rs Is Nothing Then
                        rs.Close()
                        rs.Dispose()
                    End If
                    If Not stmt Is Nothing Then
                        stmt.Dispose()
                    End If
                End Try
            End Sub

            Private Function ReplaceAccent(ByVal str As String) As String
                str = Replace(str, "á", "a")
                str = Replace(str, "à", "a")
                str = Replace(str, "â", "a")
                str = Replace(str, "ä", "a")
                str = Replace(str, "ã", "a")
                str = Replace(str, "é", "e")
                str = Replace(str, "è", "e")
                str = Replace(str, "ê", "e")
                str = Replace(str, "ë", "e")
                str = Replace(str, "í", "i")
                str = Replace(str, "ì", "i")
                str = Replace(str, "î", "i")
                str = Replace(str, "ï", "i")
                str = Replace(str, "ó", "o")
                str = Replace(str, "ò", "o")
                str = Replace(str, "ô", "o")
                str = Replace(str, "ö", "o")
                str = Replace(str, "õ", "o")
                str = Replace(str, "ú", "u")
                str = Replace(str, "ù", "u")
                str = Replace(str, "û", "u")
                str = Replace(str, "ü", "u")
                Return str
            End Function

            Public Sub CloseInternalConnection()
                If Not Me.dbConnection Is Nothing Then
                    Me.dbConnection.Close()
                    Me.dbConnection.Dispose()
                End If
            End Sub
        End Class

        Private Class MainAndIndexFileJuggler
            Private mainFileLength As Integer = -1
            Private indexFileLength As Integer = -1
            Private boundingBoxXMin As Double = Double.MaxValue
            Private boundingBoxYMin As Double = Double.MaxValue
            Private boundingBoxXMax As Double = Double.MinValue
            Private boundingBoxYMax As Double = Double.MinValue

            Private tableName As String = Nothing
            Private pathToContainingFolder As String = Nothing  'le chemin vers le dossier qui contiendra les fichiers Shape
            Private internalMainTemporaryStream As BinaryWriter = Nothing
            Private internalMainStream As BinaryWriter = Nothing
            Private internalIndexStream As BinaryWriter = Nothing
            Private counterOfCurrentRecord As Integer = 0
            Private containerOfLengthsForIndexFile As List(Of Integer)

            Public Sub New(ByVal tableName As String, ByVal pathToMainFile As String)
                Dim localPathBuffer As String = Nothing

                Me.mainFileLength = 50 'La longueur du header en mots de 16 bit
                Me.indexFileLength = 50
                Me.containerOfLengthsForIndexFile = New List(Of Integer)

                If tableName.Length > 8 Then
                    Me.tableName = tableName.Substring(0, 8).ToUpper() 'Les fichiers respectent le format 8.3 (nom.extension)
                Else
                    Me.tableName = tableName
                End If
                If pathToMainFile Is Nothing Then
                    'si le parametre Path n'a pas été renseigné, les fichiers seront placées dans le dossier courrant
                    Me.pathToContainingFolder = System.Environment.CurrentDirectory
                Else
                    Me.pathToContainingFolder = pathToMainFile
                End If

                'fichier temporaire qui sera copié à la fin dans le fichier principal (.shp)
                localPathBuffer = Me.pathToContainingFolder & "\" & Me.tableName.Replace(".", "_") & "._hp"
                If File.Exists(localPathBuffer) Then
                    File.Delete(localPathBuffer)
                End If
                Me.internalMainTemporaryStream = New BinaryWriter(File.OpenWrite(localPathBuffer))

                localPathBuffer = Me.pathToContainingFolder & "\" & Me.tableName.Replace(".", "_") & ".shp"
                If File.Exists(localPathBuffer) Then
                    File.Delete(localPathBuffer)
                End If
                Me.internalMainStream = New BinaryWriter(File.OpenWrite(localPathBuffer))

                localPathBuffer = Me.pathToContainingFolder & "\" & Me.tableName.Replace(".", "_") & ".shx"
                If File.Exists(localPathBuffer) Then
                    File.Delete(localPathBuffer)
                End If
                Me.internalIndexStream = New BinaryWriter(File.OpenWrite(localPathBuffer))
            End Sub

            Private Sub WriteLittleEndian(ByVal toWrite As Integer, ByRef streamToWrite As BinaryWriter)
                streamToWrite.Write(toWrite)
            End Sub

            Private Sub WriteLittleEndian(ByVal toWrite As Double, ByRef streamToWrite As BinaryWriter)
                streamToWrite.Write(toWrite)
            End Sub

            Private Sub WriteBigEndian(ByVal littleEndianToWrite As Integer, ByRef streamToWrite As BinaryWriter)
                Dim localArrayOfBytes As Byte()
                Dim i As Integer = -1

                localArrayOfBytes = BitConverter.GetBytes(littleEndianToWrite)
                For i = localArrayOfBytes.Length To 1 Step -1
                    streamToWrite.Write(CType(localArrayOfBytes(i - 1), Byte))
                Next
            End Sub

            Private Sub WriteBigEndian(ByVal toWrite As Double, ByRef streamToWrite As BinaryWriter)
                Dim localArrayOfBytes As Byte()
                Dim i As Integer = -1

                localArrayOfBytes = BitConverter.GetBytes(toWrite)
                For i = localArrayOfBytes.Length To 0 Step -1
                    streamToWrite.Write(CType(localArrayOfBytes(i - 1), Byte))
                Next
            End Sub

            Public Function WriteGeometryToFile(ByVal geometry As FeatureGeometry) As Boolean
                Dim localMultiPolygon As MultiPolygon = Nothing
                Dim localMultiPolygonIterator As IPolygonEnumerator = Nothing
                Dim localRingEnumerator As ICurveSegmentEnumerator = Nothing
                Dim polygonToSave As Polygon = Nothing
                Dim counterOfRings As Integer = 0
                Dim counterOfPoints As Integer = 0
                Dim byteContainerOfFirstsPointsIndex As List(Of Integer)
                Dim byteContainerOfPoints As List(Of DPoint)
                Dim externalRingToSave As Ring = Nothing
                Dim localCurveSegment As CurveSegment = Nothing
                Dim tblPoints As List(Of DPoint)
                Dim listOfInternalPolygons As IRingEnumerator = Nothing
                Dim internalRingToSave As Ring = Nothing
                Dim counterOfContentLength As Integer = -1
                Dim i As Integer = -1

                Console.WriteLine("WriteGeometryToFile ================================================================")

                ' ne gère pas les points
                If geometry.Type = GeometryType.Point _
                Or geometry.Type = GeometryType.MultiPoint Then
                    Return False
                End If

                ' modifier les bounds
                Me.boundingBoxXMin = Math.Min(Math.Min(boundingBoxXMin, geometry.Bounds.x1), geometry.Bounds.x2)
                Console.WriteLine("boundingBoxXMin = {0}", boundingBoxXMin)
                Me.boundingBoxXMax = Math.Max(Math.Max(boundingBoxXMax, geometry.Bounds.x1), geometry.Bounds.x2)
                Console.WriteLine("boundingBoxXMax = {0}", boundingBoxXMax)
                Me.boundingBoxYMin = Math.Min(Math.Min(boundingBoxYMin, geometry.Bounds.y1), geometry.Bounds.y2)
                Console.WriteLine("boundingBoxYMin = {0}", boundingBoxYMin)
                Me.boundingBoxYMax = Math.Max(Math.Max(boundingBoxYMax, geometry.Bounds.y1), geometry.Bounds.y2)
                Console.WriteLine("boundingBoxYMax = {0}", boundingBoxYMax)

                ' écrire les géometries
                byteContainerOfFirstsPointsIndex = New List(Of Integer)
                byteContainerOfPoints = New List(Of DPoint)

                Try
                    localMultiPolygon = CType(geometry, MultiPolygon) 'recuperer un ensemble de polygons
                Catch ex As Exception
                    Return False
                End Try

                localMultiPolygonIterator = localMultiPolygon.GetEnumerator()

                While localMultiPolygonIterator.MoveNext() 'recuperer polygon par polygon
                    polygonToSave = localMultiPolygonIterator.Current
                    Console.WriteLine("polygonToSave.ControlPointTotal = {0}", polygonToSave.ControlPointTotal)
                    externalRingToSave = polygonToSave.Exterior 'recuperer l'anneau exterieur
                    Console.WriteLine("externalRingToSave.ControlPointCount = {0}", externalRingToSave.ControlPointCount)
                    localRingEnumerator = externalRingToSave.GetEnumerator()

                    'TODO : mettre cette boucle dans une fonction, car elle se repete ...
                    ' ... attention au sens horlogé et anti-horlogé

                    ' traitement anneau exterieur
                    While localRingEnumerator.MoveNext()
                        localCurveSegment = localRingEnumerator.Current
                        Console.WriteLine("localCurveSegment.ControlPointCount = {0}", localCurveSegment.ControlPointCount)
                        counterOfRings = counterOfRings + 1
                        'ajouter dans le tableau d'index de points, l'index du premier point
                        byteContainerOfFirstsPointsIndex.Add(counterOfPoints)
                        counterOfPoints += localCurveSegment.ControlPointCount
                        'Verifier, eventuellement changer le sens de parcours pour le sens horlogé
                        tblPoints = Me.CorrectClockwise(localCurveSegment.SamplePoints(), True)
                        Console.WriteLine("tblPoints = ")
                        For Each p As DPoint In tblPoints
                            Console.WriteLine("     {0}; {1}", p.x, p.y)
                        Next
                        'Ecrire les points dans un tableau de bytes et incrementer le compteur "Content Length"
                        byteContainerOfPoints.AddRange(tblPoints)
                    End While

                    'Recuperer l'ensemble d'anneaux intérieurs, avec un iterateur
                    listOfInternalPolygons = polygonToSave.GetInteriorEnumerator()
                    'Parcurir la liste d'anneaux interieurs
                    While listOfInternalPolygons.MoveNext()
                        internalRingToSave = listOfInternalPolygons.Current 'recuperer l'un des anneaux interieurs
                        Console.WriteLine("internalRingToSave.ControlPointCount = {0}", internalRingToSave.ControlPointCount)
                        localRingEnumerator = internalRingToSave.GetEnumerator()
                        ' traitement anneau interieur
                        While localRingEnumerator.MoveNext()
                            localCurveSegment = localRingEnumerator.Current
                            counterOfRings = counterOfRings + 1
                            'ajouter dans le tableau d'index de points, l'index du premier point
                            byteContainerOfFirstsPointsIndex.Add(counterOfPoints)
                            counterOfPoints += localCurveSegment.ControlPointCount
                            'Verifier, eventuellement changer le sens de parcours pour le sens anti-horlogé
                            tblPoints = Me.CorrectClockwise(localCurveSegment.SamplePoints(), False)
                            Console.WriteLine("tblPoints = ")
                            For Each p As DPoint In tblPoints
                                Console.WriteLine("     {0}; {1}", p.x, p.y)
                            Next
                            'Ecrire les points dans un tableau de bytes et incrementer le compteur "Content Length"
                            byteContainerOfPoints.AddRange(tblPoints)
                        End While
                    End While

                End While

                Console.WriteLine("counterOfPoints = {0}", counterOfPoints)
                Console.WriteLine("counterOfRings = {0}", counterOfRings)
                Console.WriteLine("byteContainerOfFirstsPointsIndex.Count = {0}", byteContainerOfFirstsPointsIndex.Count)
                Console.WriteLine("byteContainerOfPoints.Count = {0}", byteContainerOfPoints.Count)

                'Ecrire le "Record Header" ("Record Number" et "Content Length", cf. specifications ESRI Shapefile)
                Me.counterOfCurrentRecord = counterOfCurrentRecord + 1
                Me.WriteBigEndian(counterOfCurrentRecord, internalMainTemporaryStream)
                Console.WriteLine("counterOfCurrentRecord = {0}", counterOfCurrentRecord)
                counterOfContentLength = 22 'car la premiere partie d'un enregistrement (Shape Type, Box, Num Parts, Num Points) ...
                ' ... contient 22 mots de 16 bits (
                counterOfContentLength += (counterOfRings * 2) 'parce que le partie "Parts" de l'enregistrement ...
                ' ... est un tableau des entiers, contenant l'index du premier point de chaque partie du milti-polygone.
                counterOfContentLength += (counterOfPoints * (2 * 4)) 'parce que la pertie "Points" de l'enregistrement ...
                ' ... est un tableau de points et chaque point est composé de 2 valeurs de type double
                Me.WriteBigEndian(counterOfContentLength, internalMainTemporaryStream)
                Console.WriteLine("counterOfContentLength = {0}", counterOfContentLength)
                Me.containerOfLengthsForIndexFile.Add(counterOfContentLength + 4)

                Me.mainFileLength += (4 + counterOfContentLength) 'Le contenu du fichier augmente avec la taille de l'enregistrement + 4 mots ...
                ' ... de 16 bits, c.-à-d. la taille de l'entête de l'enregistrement
                Me.indexFileLength += 4 'Le contenu du fichier augmente avec la taille de l'enregistrement

                'Ajouter "Shape Type" (5 => multi-polygone cf. specifications ESRI Shapefile)
                Me.WriteLittleEndian(5, internalMainTemporaryStream)

                'Ajouter "Bounding Box" dans l'ordre : Xmin, Ymin, Xmax, Ymax
                Me.WriteLittleEndian(Math.Min(geometry.Bounds.x1, geometry.Bounds.x2), internalMainTemporaryStream)
                Console.WriteLine("Math.Min(geometry.Bounds.x1, geometry.Bounds.x2) = {0}", Math.Min(geometry.Bounds.x1, geometry.Bounds.x2))
                Me.WriteLittleEndian(Math.Min(geometry.Bounds.y1, geometry.Bounds.y2), internalMainTemporaryStream)
                Console.WriteLine("Math.Min(geometry.Bounds.y1, geometry.Bounds.y2) = {0}", Math.Min(geometry.Bounds.y1, geometry.Bounds.y2))
                Me.WriteLittleEndian(Math.Max(geometry.Bounds.x1, geometry.Bounds.x2), internalMainTemporaryStream)
                Console.WriteLine("Math.Max(geometry.Bounds.x1, geometry.Bounds.x2) = {0}", Math.Max(geometry.Bounds.x1, geometry.Bounds.x2))
                Me.WriteLittleEndian(Math.Max(geometry.Bounds.y1, geometry.Bounds.y2), internalMainTemporaryStream)
                Console.WriteLine("Math.Max(geometry.Bounds.y1, geometry.Bounds.y2) = {0}", Math.Max(geometry.Bounds.y1, geometry.Bounds.y2))

                'Ajouter "Num Parts" (nombre d'anneaux dans le multi-polygone)
                Me.WriteLittleEndian(counterOfRings, internalMainTemporaryStream)
                Console.WriteLine("counterOfRings = {0}", counterOfRings)

                'Ajouter "Num Points" (nombre de points dans le polygone. Attention, un point est composé de deux dimessions)
                Me.WriteLittleEndian(counterOfPoints, internalMainTemporaryStream)
                Console.WriteLine("counterOfPoints = {0}", counterOfPoints)

                'Ajouter "byteContainerOfFirstsPointsIndex" (index contenant le premier point de chaque anneau)
                For i = 0 To byteContainerOfFirstsPointsIndex.Count - 1
                    Me.internalMainTemporaryStream.Write(byteContainerOfFirstsPointsIndex(i))
                    Console.WriteLine("byteContainerOfFirstsPointsIndex({0}) = {1}", i, byteContainerOfFirstsPointsIndex(i))
                Next

                'Ajouter "byteContainerOfPoints" (les données des points
                For i = 0 To byteContainerOfPoints.Count - 1
                    Me.internalMainTemporaryStream.Write(byteContainerOfPoints(i).x)
                    Me.internalMainTemporaryStream.Write(byteContainerOfPoints(i).y)
                    Console.WriteLine("byteContainerOfPoints({0}).x = {1}", i, byteContainerOfPoints(i).x)
                    Console.WriteLine("byteContainerOfPoints({0}).y = {1}", i, byteContainerOfPoints(i).y)
                Next

                Return True
            End Function

            Public Function WriteGeometryToFile(ByVal geometryContainer As Feature) As Boolean
                Return Me.WriteGeometryToFile(geometryContainer.Geometry)
            End Function

            Private Function CorrectClockwise(ByVal tableToOrdonnate() As DPoint, ByVal clockWiseOrder As Boolean) As List(Of DPoint)
                Dim counterFinal As Integer = 0
                Dim resultatPartiel As Double = -1
                Dim indexFirstPoint, indexSecondPoint, indexThirdPoint, verticesNumber As Integer
                Dim aRetourner As List(Of DPoint)
                Dim i As Integer = -1

                verticesNumber = tableToOrdonnate.Length - 1
                aRetourner = New List(Of DPoint)
                For i = 0 To tableToOrdonnate.Length - 1
                    aRetourner.Add(tableToOrdonnate(i))
                Next

                ' verifier si le polygone est parcouru dans le sens horlogé
                For i = 0 To verticesNumber
                    indexFirstPoint = i
                    indexSecondPoint = (i + 1) Mod verticesNumber
                    indexThirdPoint = (i + 2) Mod verticesNumber

                    resultatPartiel = _
                        (tableToOrdonnate(indexSecondPoint).x - tableToOrdonnate(indexFirstPoint).x) * _
                        (tableToOrdonnate(indexThirdPoint).y - tableToOrdonnate(indexSecondPoint).y) - _
                        (tableToOrdonnate(indexSecondPoint).y - tableToOrdonnate(indexFirstPoint).y) * _
                        (tableToOrdonnate(indexThirdPoint).x - tableToOrdonnate(indexSecondPoint).x)
                    If resultatPartiel < 0 Then
                        counterFinal = counterFinal - 1
                    End If
                    If resultatPartiel > 0 Then
                        counterFinal = counterFinal + 1
                    End If
                Next

                'Si counterFinal < 0 => le polygone est parcouru en sens horlogé
                If (counterFinal < 0) Xor clockWiseOrder Then
                    aRetourner.Reverse()
                End If

                Return aRetourner
            End Function

            Public Sub FinaliseAndCloseFiles()
                Dim localPathToMainTemporaryFile As String = Nothing
                Dim localMainReader As BinaryReader = Nothing
                Dim localMainBuffer As Byte
                Dim currentOffset As Integer = -1
                Dim i As Integer = -1

                ' composition fichier principal
                ' écrire le header dans le fichier principal
                Me.WriteBigEndian(9994, internalMainStream) '/File code (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalMainStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalMainStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalMainStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalMainStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalMainStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(mainFileLength, internalMainStream) 'File Length (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(1000, internalMainStream) 'Version (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(5, internalMainStream) 'Shape Type : 5<=>Polygone (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxXMin, internalMainStream) 'Xmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxYMin, internalMainStream) 'Ymin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxXMax, internalMainStream) 'Xmax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxYMax, internalMainStream) 'Ymax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalMainStream) 'Zmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalMainStream) 'Zmax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalMainStream) 'Mmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalMainStream) 'Mmax (cf. ESRI Shapefile Technical Description)

                ' copier le fichier temporaire dans le fichier principal
                Me.internalMainTemporaryStream.Flush()

                ' Fermer le fichier temporaire (actuellement ouvert en ecriture)
                Me.internalMainTemporaryStream.Close()

                localPathToMainTemporaryFile = Me.pathToContainingFolder & "\" & Me.tableName.Replace(".", "_") & "._hp"
                localMainReader = New BinaryReader(File.OpenRead(localPathToMainTemporaryFile)) 'Ouvrir le fichier temporaire en lecture

                Try
                    While True
                        localMainBuffer = localMainReader.ReadByte()
                        Me.internalMainStream.Write(localMainBuffer)
                    End While
                Catch ex As Exception
                End Try

                localMainReader.Close()

                'Début composition fichier index
                'Début écrire le header dans le fichier index
                Me.WriteBigEndian(9994, internalIndexStream) 'File code (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalIndexStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalIndexStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalIndexStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalIndexStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(0, internalIndexStream) 'Unused (cf. ESRI Shapefile Technical Description)
                Me.WriteBigEndian(indexFileLength, internalIndexStream) 'File Length (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(1000, internalIndexStream) 'Version (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(5, internalIndexStream) 'Shape Type : 5<=>Polygone (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxXMin, internalIndexStream) 'Xmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxYMin, internalIndexStream) 'Ymin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxXMax, internalIndexStream) 'Xmax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(boundingBoxYMax, internalIndexStream) 'Ymax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalIndexStream) 'Zmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalIndexStream) 'Zmax (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalIndexStream) 'Mmin (cf. ESRI Shapefile Technical Description)
                Me.WriteLittleEndian(0.0, internalIndexStream) 'Mmax (cf. ESRI Shapefile Technical Description)

                ' ecrire les données dans le fichier index
                currentOffset = 50
                For i = 0 To containerOfLengthsForIndexFile.Count - 1
                    Me.WriteBigEndian(currentOffset, internalIndexStream)
                    Me.WriteBigEndian(containerOfLengthsForIndexFile(i), internalIndexStream)
                    currentOffset += containerOfLengthsForIndexFile(i)
                Next

                Me.internalMainStream.Close()
                Me.internalIndexStream.Close()

                If File.Exists(localPathToMainTemporaryFile) Then
                    File.Delete(localPathToMainTemporaryFile)
                End If
            End Sub

        End Class

    End Class

End Namespace
